package org.netkernel.http.transport;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.mortbay.jetty.HttpHeaders;
import org.mortbay.jetty.HttpMethods;
import org.netkernel.container.IKernel;
import org.netkernel.http.rep.IHttpRedirect;
import org.netkernel.http.transport.representation.HTTPRequestResponseRepresentation;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.EndpointMetaImpl;
import org.netkernel.layer0.meta.impl.SimpleSpaceElementsImpl;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.urii.ActiveURIEncoding;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.RegexRewriterRule;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.IStandardOverlay;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.module.standard.endpoint.TransparentOverlayImpl;
import org.netkernel.module.standard.endpoint.ConfiguredOverlayImpl.IConfig;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.impl.NetKernelError;
import org.w3c.dom.Document;
import org.w3c.dom.Node;


public class HTTPBridgeOverlayConfig implements IConfig
{
	private final RegexRewriterRule[] mRules;
	private final ISpace mDelegateSpace;
	private final String mExceptionHandler;
	private final IEndpoint mEndpoint;
	private final String mElementId;
	private final EndpointMetaImpl mMeta;
	private String mPostParamEncoding="UTF-8";
	
	public static final String HTTP_CACHE_EXPIRES_DELTA_SECONDS="HTTP_CACHE_EXPIRES_DELTA_SECONDS";
	public static final String HTTPBRIDGE_TRANSPARENT_PASS_THROUGH="HTTPBRIDGE_TRANSPARENT_PASS_THROUGH";
	
	
	public HTTPBridgeOverlayConfig(ISpace aDelegateSpace,ISpaceMeta aDelegateMeta, INKFRequestContext aContext, IEndpoint aEndpoint) throws Exception
	{	mDelegateSpace=aDelegateSpace;
		mExceptionHandler=(String)aContext.getParamValue("exceptionHandler");
		mEndpoint=aEndpoint;
		Document config=null;
		if(aContext.getParamValue("config")!=null)
		{	config=(Document)aContext.source("param:config",Document.class);
		}
		IGrammarGroup grammarRoot = BNFGrammarFactory.createChoiceGroup(null);
		if(config!=null)
		{	XMLReadable c=new XMLReadable(config);
			List<RegexRewriterRule> rules = new ArrayList<RegexRewriterRule>();
			for (Iterator<Node> i=c.getNodes("/config/rewrite").iterator(); i.hasNext(); )
			{	XMLReadable rewrite=new XMLReadable(i.next());
				String match=rewrite.getText("match");
				String to=rewrite.getText("to");
				RegexRewriterRule rule = new RegexRewriterRule(match,to);
				rules.add(rule);
				BNFGrammarFactory.createRegex(grammarRoot, match);
			}
			mRules = new RegexRewriterRule[rules.size()];
			rules.toArray(mRules);
			String encoding=c.getText("/config/defaultPostParamEncoding");
			if (encoding.length()>0)
			{	mPostParamEncoding=encoding;
			}
		}
		else
		{	mRules = new RegexRewriterRule[1];
			mRules[0]=new RegexRewriterRule("(htt(p|ps)://.*)", "$1");
		}
		
		//setup meta
		mElementId = Utils.generateUniqueEndpointIdentifier(mDelegateSpace, aEndpoint, "HTTPBridgeOverlayConfig");
		IIdentifierGrammar grammar=BNFGrammarFactory.createGrammar(grammarRoot, aContext.getKernelContext().getKernel().getLogger());
		mMeta = new EndpointMetaImpl("HTTPBridge",null,null,null,IRequest.VERB_SOURCE,grammar,EndpointMetaImpl.NO_ARGS,EndpointMetaImpl.NO_RESPONSES,EndpointMetaImpl.NO_EXCEPTIONS);
		mMeta.addAdditionalField("private", true);
	}
	
	String getDefaultPostParamEncoding()
	{	return mPostParamEncoding;
	}

	public void onRequest(String aID, INKFRequestContext context) throws Exception
	{		//Pass through mode
			if(context.getThisRequest().getHeaderValue(HTTPBRIDGE_TRANSPARENT_PASS_THROUGH)!=null)
			{	//Perform any inner request rewriting
				String innerID=context.getThisRequest().getIdentifier();
				innerID = Utils.decode(innerID);
				if(mRules.length>0)
				{	for(RegexRewriterRule rule : mRules )
					{	if(rule.matches(innerID))
						{	innerID=rule.map(innerID);
						}
					}
				}
				//Prepare sub-request into wrapped space only - no HTTP bridge space.
				INKFRequest req=context.createRequest(innerID);
				req.setRepresentationClass(IBinaryStreamRepresentation.class);
				context.createResponseFrom(context.issueRequestForResponse(req));
				return;
			}
		
			//Normal mode
			HTTPRequestResponseRepresentation httpRepresentation = (HTTPRequestResponseRepresentation)context.getThisRequest().getPrimary();
			//Prepare a root request context.
			if(httpRepresentation==null)
			{	throw new Exception("HTTPTransport Fulcrum space is misconfigured and is not providing the HTTPRequestResponseRepresentation to the HTTPBridge");				
			}
			HttpServletRequest httpRequest = httpRepresentation.getHttpRequest();
			HTTPRequestSpace httpRequestSpace = new HTTPRequestSpace(httpRequest,this);
			IRequestScopeLevel injectionScope = RequestScopeLevelImpl.createDurableScopeLevel(httpRequestSpace, context.getKernelContext().getThisKernelRequest().getRequestScope());
			injectionScope=RequestScopeLevelImpl.createNonDurableScopeLevel(mDelegateSpace, injectionScope);
			
			//Perform any inner request rewriting
			String innerID=context.getThisRequest().getIdentifier();
			innerID = Utils.decode(innerID);
			if(mRules.length>0)
			{	for(RegexRewriterRule rule : mRules )
				{	if(rule.matches(innerID))
					{	innerID=rule.map(innerID);
					}
				}
			}
			//Prepare sub-request into prepared context
			INKFRequest req=context.createRequest(innerID);
			req.setRequestScope(injectionScope);
			req.setRepresentationClass(IBinaryStreamRepresentation.class);
			
			HttpServletResponse httpResponse = httpRepresentation.getHttpResponse();
			try
			{	//Commission the prepared context
				httpRequestSpace.onCommissionSpace(context.getKernelContext().getKernel());
				////////////////////////
				//Issue inner sub-request
				////////////////////////
				INKFResponseReadOnly innerResponse = context.issueRequestForResponse(req);
				if(!httpRepresentation.needsHandling())
				{	//Response has already been written by the HTTP request handler timeout code.
					INKFResponse response=context.createResponseFrom(null);
					response.setExpiry(INKFResponse.EXPIRY_ALWAYS);
					return; 					
				}
				IBinaryStreamRepresentation innerRepresentation = (IBinaryStreamRepresentation)innerResponse.getRepresentation();
				httpRequestSpace.applyResponseMetadata(innerResponse.getHeaders(),context);
				
				if(innerResponse.hasHeader(IHttpRedirect.HTTP_REDIRECT_METADATA_IDENTIFIER))
				{	//Handle Redirects
					IHttpRedirect redirect=(IHttpRedirect)innerResponse.getHeader(IHttpRedirect.HTTP_REDIRECT_METADATA_IDENTIFIER);
					httpResponse.setStatus(redirect.getType());
					httpResponse.setHeader(HttpHeaders.LOCATION, redirect.getURL());
					if(redirect.getType()==IHttpRedirect.HTTP_REDIRECT_300_MULTIPLE_CHOICES)
					{	//Write choices...
						httpResponse.setContentType("text/plain");
						OutputStreamWriter w=new OutputStreamWriter(httpResponse.getOutputStream());
						for(String s: redirect.getChoices())
						{	w.write(s);
							w.write("\n");
						}
						w.close();
					}
					else if(innerRepresentation!=null)
					{	//Comply with HTTP1.1 spec and send short note as body if it is present.
						writeStream(innerRepresentation, httpResponse, innerResponse.getMimeType(), httpRepresentation.getHttpRequest());					
					}
				}
				else
				{	////////////////////////////////////////
					//Process HTTP Response
					////////////////////////////////////////
					boolean writestream=writeHeaders(innerResponse, httpResponse, httpRequestSpace);
					if (innerRepresentation!=null && writestream)
					{	//Etag Processing
						String method=httpRequest.getMethod();
						if((method.equals(HttpMethods.GET ) || method.equals(HttpMethods.HEAD)) && !innerResponse.isExpired())
						{	String etag=httpRequest.getHeader(HttpHeaders.IF_NONE_MATCH);  //Doesn't matter if null
							String tagString="\""+Integer.toHexString(innerRepresentation.hashCode())+"\"";
							//Write the Etag
							httpResponse.setHeader(HttpHeaders.ETAG,tagString);
							if(tagString.equals(etag))
							{	//Not modified
								writestream=false;
								httpResponse.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
							}
						}
						//Write the representation.
						if(writestream && !method.equals(HttpMethods.HEAD))
						{	httpResponse.setDateHeader(HttpHeaders.LAST_MODIFIED, System.currentTimeMillis());
							writeStream(innerRepresentation, httpResponse, innerResponse.getMimeType(), httpRequest);
						}
					}
					else if (writestream)
					{	throw context.createFormattedException("EX_NULL_RESPONSE", "MSG_NULL_RESPONSE", null, null);
					}
				}
			}
			catch(org.mortbay.jetty.EofException e)
			{	//Requestor has gone away so no point handling Exception.
				context.logRaw(INKFRequestContext.LEVEL_FINE, e.getMessage());
				context.logRaw(INKFRequestContext.LEVEL_FINER, e.toString());
			}
			catch (NKFException e)
			{	handleException(e,context,injectionScope,httpResponse,httpRequest,httpRequestSpace);
			}
			catch (NetKernelError e)
			{	handleException(e,context,injectionScope,httpResponse,httpRequest,httpRequestSpace);
			}
			catch (Exception e)
			{	INetKernelException ex = context.createFormattedException("EX_IN_HTTP_BRIDGE",null,null,e);
				handleException(ex,context,injectionScope,httpResponse,httpRequest,httpRequestSpace);
			}
			INKFResponse response=context.createResponseFrom(null);
			response.setExpiry(INKFResponse.EXPIRY_ALWAYS);
	}
	
	private boolean writeHeaders(INKFResponseReadOnly innerResponse, HttpServletResponse httpResponse, HTTPRequestSpace httpRequestSpace) throws IOException, NKFException
	{
		boolean writestream=true;
		
		httpResponse.setHeader(HttpHeaders.SERVER,  HTTPTransportEndpoint.getServerHeader() );
		
		//Process Cookies
		Map cookies=httpRequestSpace.getResponseCookies();
		Iterator i=cookies.keySet().iterator();
		while(i.hasNext())
		{		Cookie c=(Cookie)cookies.get(i.next());
				httpResponse.addCookie(c);
		}
		
		//Process Headers
		Map headers=httpRequestSpace.getResponseHeaders();
		i=headers.keySet().iterator();
		while(i.hasNext())
		{	String name=(String)i.next();
			Object value=headers.get(name);
			if (value instanceof Long)
			{	httpResponse.setDateHeader(name, (Long)value);
			}
			else if (value instanceof Integer)
			{	httpResponse.setIntHeader(name, (Integer)value);
			}
			else
			{	httpResponse.setHeader(name, value.toString());
			}
		}
		
		if(innerResponse.hasHeader(HTTP_CACHE_EXPIRES_DELTA_SECONDS))
		{	String vs=innerResponse.getHeader(HTTP_CACHE_EXPIRES_DELTA_SECONDS).toString();
			try
			{	httpResponse.setDateHeader(HttpHeaders.EXPIRES, System.currentTimeMillis()+Long.parseLong(vs)*1000L);
			}
			catch(Exception e)
			{	e.printStackTrace();  //Wrong number format.				
			}
			httpResponse.setHeader(HttpHeaders.CACHE_CONTROL, "public, max-age="+vs);
		}

		//Process Response Code
		httpResponse.setStatus(httpRequestSpace.getResponseCode());
		
		//Process Redirect
		if(httpRequestSpace.shouldRedirect())
		{	httpResponse.sendRedirect(httpRequestSpace.getResponseRedirect());
			writestream=false;
		}
		return writestream;
	}
	
	private void writeStream(IBinaryStreamRepresentation innerRepresentation, HttpServletResponse httpResponse, String mimeType, HttpServletRequest httpRequest) throws Exception
	{	//String mimeType=innerResponse.getMimeType();
		String encoding=innerRepresentation.getEncoding();
		if(encoding!=null)
		{	int gp=encoding.indexOf("gzip");
			if(gp>=0)
			{	httpResponse.setHeader("Content-Encoding", "gzip");
				if(gp>0)
				{	encoding=encoding.substring(0,gp-1);  	//Strip +gzip tag							
				}
				else
				{	encoding=null;  											//Prevent writing gzip encoding on mimetype header.							
				}
			}
		}
		if (mimeType!=null)
		{	//Special case to make XHTML work with crappy Internet Explorer
			if(mimeType.equals("application/xhtml+xml"))
			{	if(httpRequest.getHeader(HttpHeaders.USER_AGENT).contains("MSIE"))
				{	mimeType="text/html";
				}
			}
			if( (encoding!=null) && mimeType.startsWith("text/") )    //Encoding when text but we need a complete mimetype - encoding database see: http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7
			{	mimeType+=";charset="+encoding;
			}
			httpResponse.setContentType(mimeType);
		}
		//Content-Length
		if (innerRepresentation instanceof IReadableBinaryStreamRepresentation)
		{	IReadableBinaryStreamRepresentation rbs = (IReadableBinaryStreamRepresentation)innerRepresentation;
			httpResponse.setContentLength(rbs.getContentLength());
		}
		
		OutputStream s=httpResponse.getOutputStream();
		innerRepresentation.write(s);
		//s.close();
	}		

	private void handleException(Throwable e, INKFRequestContext context, IRequestScopeLevel injectionContext, HttpServletResponse httpResponse, HttpServletRequest httpRequest, HTTPRequestSpace httpRequestSpace) throws Exception
	{	try
		{	httpResponse.reset();
		}
		catch(IllegalStateException ies)
		{	/*Can't reset the response - already committed*/
			context.logRaw(INKFRequestContext.LEVEL_FINE, "Exception but HTTPResponse is complete cannot be reset. "+e.getMessage());
			return;
		}
		try
		{
			//explicit serialisation so that handler considers exception as data
			INKFRequest req=context.createRequest("literal:exception");
			req.setVerb(INKFRequestReadOnly.VERB_TRANSREPRESENT);
			req.addPrimaryArgument(e);
			req.setRepresentationClass(IBinaryStreamRepresentation.class);
			req.setRequestScope(injectionContext);
			Object serializedException=context.issueRequest(req);
			
			if (mExceptionHandler!=null)
			{	req=context.createRequest(mExceptionHandler);
				req.setRequestScope(injectionContext);
				req.addArgumentByValue("operand",serializedException);
				req.setRepresentationClass(IBinaryStreamRepresentation.class);
				INKFResponseReadOnly<IBinaryStreamRepresentation> response = context.issueRequestForResponse(req);
				httpRequestSpace.applyResponseMetadata(response.getHeaders(),context);
				boolean writestream=writeHeaders(response, httpResponse, httpRequestSpace);
				if (writestream)
				{	writeStream(response.getRepresentation(), httpResponse, response.getMimeType(), httpRequest);
				}
			}
			else
			{	httpResponse.setContentType("text/xml");
				OutputStream s=httpResponse.getOutputStream();
				IBinaryStreamRepresentation result = (IBinaryStreamRepresentation)serializedException;
				result.write(s);
				s.close();
			}
		}
		catch (Exception e2)
		{	context.logRaw(INKFRequestContext.LEVEL_WARNING, e2.toString());
			if (e instanceof INetKernelException)
			{	throw (INetKernelException)e;
			}
			else if (e instanceof NetKernelError)
			{	throw (NetKernelError)e;
			}
			else throw e2;
		}
	}

	public IMetaRepresentation getElementMeta(String aElementId)
	{	return mMeta;
	}

	public ISpaceElements getElements(INKFRequestContext aContext) 	throws Exception
	{	ISpaceElements result=new SimpleSpaceElementsImpl(new SimpleIdentifierImpl(mElementId),mEndpoint);
		return result;
	}

	public String getResolvedElement(INKFRequestReadOnly aRequest) throws Exception
	{	String result=null;
		String id=aRequest.getIdentifier();
		for(RegexRewriterRule rule : mRules)
		{	if(rule.matches(id))
			{	result=mElementId;
				break;
			}
		}
		return result;
	}

	public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception
	{	return IStandardEndpoint.NO_SPACES;
	}
}